#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

import errno, os, sys, datetime, hashlib, paramiko, re, requests, ssl, socket, time, threading, traceback, urllib2, urlparse, uuid, copy

from defines import *
from httplib import BadStatusLine
from packaging import version
from pyVmomi import vim, vmodl
from pyVim.task import WaitForTask, TaskBlocked
from pyVim.connect import SmartConnect, Disconnect

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

import logging
import logging.handlers
logger = logging.getLogger(VM_LOGGER_NAME)

# load vix error text
vixErrorText = {}
errorsFilePath = '/var/packages/ActiveBackup/target/lib/vddk65/errors.html'

with open(errorsFilePath) as file:
    content = file.read()

matches = re.findall(r'<dt.*?>(\d+).*?<dd>(.*?)</dd>', content, re.DOTALL)

for match in matches:
    vixErrorText[match[0]] = match[1]

del matches, content

def isSuccessStatusCode(code):
    return 200 <= code and code < 300

def _url_fix(url, charset='utf-8'):
    if isinstance(url, unicode):
        url = url.encode(charset, 'ignore')
    return urllib2.quote(url)

def _query_fix(qs, charset='utf-8'):
    if isinstance(qs, unicode):
        qs = qs.encode(charset, 'ignore')
    qs = qs.replace('%', '%25')
    qs = qs.replace('#', '%23')
    qs = qs.replace('&', '%26')
    qs = qs.replace('=', '%3D')
    qs = qs.replace('?', '%3F')
    qs = qs.replace('+', '%2B')
    return qs

class GuestNotEnoughSpace(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class GuestOSException(Exception):
    def __init__(self, address, errno, msg):
        self.address = address
        self.errno = errno
        self.msg = msg
    def __str__(self):
        return self.msg
class GuestOSMetaControlFail(Exception):
    def __init__(self, address, errno, msg, miss_attr_list):
        self.address = address
        self.errno = errno
        self.msg = msg
        self.miss_attr_list = miss_attr_list
    def __str__(self):
        return self.msg
class LocalStatFail(Exception):
    def __init__(self, errno, msg):
        self.errno = errno
        self.msg = msg
    def __str__(self):
        return self.msg

class ErrCertVerified(Exception):
    def __init__(self, msg, cert):
        self.msg = msg
        self.cert = cert
    def __str__(self):
        return self.msg

class ErrUnauth(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class Main:
    def __init__(self, hostname, username, password, thumbprint, port, protocol, taskPool):
        self.si = None
        self.task_pool = taskPool
        self.api_type = None
        self.api_version = None
        self.auth_info = {
            'hostname'  : hostname,
            'username'  : username,
            'password'  : password,
            'thumbprint': thumbprint,
            'port'      : port
        }

        self.always_trust_cert = False
        self.lock = threading.Lock()
        self.lockNum = 0
        self.lockNumLock = threading.Lock()
        self.error = None
        self.failed_count = 0

    def __del__(self):
        Disconnect(self.si)
        logger.info('disconnect %s' % self.auth_info['hostname'])

    def dispatch(self, command, argument):
        command = getattr(self, command)
        argument = argument or {}
        response = {}
        always_trust_cert = False
        force_connect = False
        if 'always_trust_cert' in argument:
            always_trust_cert = argument['always_trust_cert']
            argument.pop('always_trust_cert')

        if 'force_connect' in argument:
            force_connect = argument['force_connect']
            argument.pop('force_connect')

        try:
            self.connect(always_trust_cert, force_connect)
            response['data'] = command(**argument)
            response['success'] = True
        except Exception as e:
            logger.error(traceback.format_exc())
            response = self._parse_exception(e)
            response['success'] = False

        return response

    def _parse_exception(self, e):
        error_class = type(e).__name__
        error_message = ''
        error_params = {}

        if type(e) == socket.error:
            if e[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE, errno.ECONNRESET):
                error_class = 'vim.fault.HostConnectFault'
            error_message = e[1]
        elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.exceptions.SSLError, requests.ConnectionError, requests.Timeout):
            error_class = 'vim.fault.HostConnectFault'
            error_message = 'Unknown connection error'
        elif type(e) == ErrCertVerified:
            error_class = 'ErrCertVerified'
            error_message = e.msg + ';' + e.cert
        elif type(e) == vmodl.fault.SystemError:
            m = re.search(r'vix error codes = \((.*?),', e.msg)
            if m != None and m.lastindex == 1:
                error_message = vixErrorText[m.group(1)]
            else:
                error_message = e.msg
        elif type(e) == GuestOSMetaControlFail:
            error_message = e.msg
            error_params['miss_attr_list'] = e.miss_attr_list
        elif hasattr(e, 'faultMessage') and len(e.faultMessage) > 0:
            if hasattr(e.faultMessage[0], 'message'):
                error_message = e.faultMessage[0].message
            else:
                error_message = e.msg

            if hasattr(e.faultMessage[0], 'arg') and \
                    len(e.faultMessage[0].arg) > 0 and \
                    e.faultMessage[0].arg[0].value == 'msg.snapshot.error-QUIESCINGERROR':
                error_class = 'vim.fault.ApplicationQuiesceFault'

        elif hasattr(e, 'msg'):
            error_message = e.msg
            if error_message.startswith('Server has wrong SHA1 thumbprint:'):
                error_class = 'ThumbprintMismatch'
        elif isinstance(e, OSError) or type(e) == socket.gaierror:
            error_message = str(e.strerror)
        else:
            error_message = str(e)

        return {
            'error_class': error_class,
            'error_message': error_message,
            'error_params': error_params,
        }

    def check_retry(self):
        return self.failed_count < 2

    def connect_failed(self):
        self.failed_count += 1
        logger.error('failed count: %s\n' % self.failed_count)

    def smart_connect(self, always_trust_cert, force_connect):
        on_disk_thumbprint = self.auth_info['thumbprint']
        context = ssl.create_default_context()
        if True == always_trust_cert:
            context = ssl._create_unverified_context()

        context.load_verify_locations(capath='/etc/ssl/certs')

        try:
            if force_connect or self.check_retry():
                si = SmartConnect(
                    host=self.auth_info['hostname'],
                    user=self.auth_info['username'],
                    pwd=self.auth_info['password'],
                    port=self.auth_info['port'],
                    sslContext=context)
                self.failed_count = 0

                return si
        except ssl.SSLError:

            cert_pem = ssl.get_server_certificate((self.auth_info['hostname'], self.auth_info['port']))
            currentThumbprint = self.getServerCert()['thumbprint']
            if currentThumbprint != on_disk_thumbprint:
                logger.error('certificate verified error!! check_hostname: %s, on_disk_thumbprint: %s, current thumbprint: %s, always_trust_cert: %d' % (self.auth_info['hostname'] , on_disk_thumbprint ,  currentThumbprint , always_trust_cert))
                raise ErrCertVerified(msg='verify_failed', cert = cert_pem)

        except ssl.CertificateError:

            cert_pem = ssl.get_server_certificate((self.auth_info['hostname'], self.auth_info['port']))
            currentThumbprint = self.getServerCert()['thumbprint']
            if currentThumbprint != on_disk_thumbprint:
                logger.error('certificate verified error!! (hostname not match) check_hostname: %s, on_disk_thumbprint: %s, current thumbprint: %s, always_trust_cert: %d' % (self.auth_info['hostname'] , on_disk_thumbprint , currentThumbprint , always_trust_cert))
                raise ErrCertVerified(msg='hostname_not_match', cert = cert_pem)

        except vim.fault.InvalidLogin as e:
            logger.error('Incorrect user/password')
            self.connect_failed()
            raise e

        # not pass cert verified
        try:
            if force_connect or self.check_retry():
                si = SmartConnect(
                    host=self.auth_info['hostname'],
                    user=self.auth_info['username'],
                    pwd=self.auth_info['password'],
                    port=self.auth_info['port'],
                    sslContext=ssl._create_unverified_context())
                self.failed_count = 0

                return si

        except vim.fault.InvalidLogin as e:
            logger.error('Incorrect user/password')
            self.connect_failed()
            raise e

        logger.info('Auto block non-force connect request during UnAuth state')
        raise ErrUnauth(msg='Within Unauthentication state')

    def _connect(self, always_trust_cert, force_connect):
        if not self.auth_info['password']:
            raise vim.fault.InvalidLogin(msg='Empty password')

        self.si = self.smart_connect(always_trust_cert, force_connect)

        self.always_trust_cert = always_trust_cert
        about = self.si.content.about
        self.api_type = about.apiType
        self.api_version = about.apiVersion

        if self._is_vcenter():
            self.instance_uuid = about.instanceUuid
        else:
            host_obj = self._get_obj_by_moid('ha-host')
            self.instance_uuid = host_obj.summary.hardware.uuid

        logger.info('connect %s' % self.auth_info['hostname'])

    def connect(self, always_trust_cert, force_connect):
        try:
            self.lockNumLock.acquire()
            self.lockNum += 1
            self.lockNumLock.release()
            self.lock.acquire()

            if self.error:
                raise self.error
            if self.si == None:
                self._connect(always_trust_cert, force_connect)
            try:
                self.si.CurrentTime()
            except vim.fault.NotAuthenticated:
                logger.error('reauth %s' % self.auth_info['hostname'])
                self._connect(always_trust_cert, force_connect)
        except Exception as e:
            self.error = e
            raise
        finally:
            self.lockNumLock.acquire()
            self.lockNum -= 1
            if self.lockNum == 0:
                self.error = None
            self.lockNumLock.release()
            self.lock.release()

    def getServerInfo(self, reconnect=False):
        def _is_shell_running(hosts):
            for host in hosts:
                for service in host.configManager.serviceSystem.serviceInfo.service:
                    if (service.key == 'TSM' or service.key == 'TSM-SSH') and service.running == False:
                        return False
            return True

        def _is_shell_connectable(hostname, port):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(10)
            try :
                sock.connect((hostname, port))
                sock.close()
            except:
                return False
            return True

        def _has_api_feature(licenses):
            for license in licenses:
                hasCapability = False
                for prop in license.properties:
                    if prop.key == 'feature' and prop.value.key == 'vstorage':
                        hasCapability = True
                if not hasCapability:
                    continue
                if license.total == 0 and license.used == 0:
                    continue
                return True
            return False

        if reconnect:
            si = self.smart_connect(self.always_trust_cert, False)
            Disconnect(si)

        data = {}
        data['api_type'] = self.api_type
        data['host_version'] = self.api_version
        data['instance_uuid'] = self.instance_uuid
        data['licensed'] = self._is_vcenter() or _has_api_feature(self.si.content.licenseManager.licenses)
        data['shell_available'] = data['licensed'] or \
                (_is_shell_running(self._get_objs([vim.HostSystem])) and \
                _is_shell_connectable(self.auth_info['hostname'], 22))
        data['guest_available'] = data['licensed'] and (int(self.api_version.split('.')[0]) >= 5)
        return data

    def getServerCert(self):
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock = context.wrap_socket(sock, server_hostname=self.auth_info['hostname'])
        sock.connect((self.auth_info['hostname'], self.auth_info['port']))
        cert = sock.getpeercert(True)
        sock.close()

        sha1 = hashlib.sha1()
        sha1.update(cert)
        digest = sha1.hexdigest().lower()

        s, i = '', 0
        for c in digest:
            s += c
            i += 1
            if i % 2 == 0:
                s += ':'

        return {'thumbprint': s[:-1]}

    def getHostVersion(self, host_id):
        host = self._get_obj_by_moid(host_id, [vim.HostSystem])
        configs = host.parent.environmentBrowser.QueryConfigOptionDescriptor()
        versions = []
        for config in configs:
            version = {}
            version['key'] = config.key.split('-')[1]
            version['support'] = config.createSupported
            version['default'] = config.defaultConfigOption
            versions.append(version)
        return versions

    def getVmInfo(self, vm_id, get_detail=True):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        return self._parse_vm_data(vm, get_detail)

    def getDatastoreInfo(self, datastore_infos):
        ret = []
        for datastore_info in datastore_infos:
            datastore = self._get_obj_by_moid(datastore_info['datastore_id'], [vim.Datastore])
            ret.append(self._parse_datastore_data(datastore))
        return ret

    def getEntityPath(self, entity=None, entityId=None, viewType=None, attr='_moId'):
        d = {'entity': entity or self._get_obj_by_moid(entityId), 'path': ''}

        def _get_datacenter_path():
            while d['entity'] != self._get_root_obj():
                d['path'] = '/' + getattr(d['entity'], attr) + d['path']
                d['entity'] = d['entity'].parent

        def _get_resource_pool_path():
            while type(d['entity'].parent) != vim.ComputeResource and \
                  type(d['entity'].parent) != vim.ClusterComputeResource:
                if self._is_vcenter():
                    d['path'] = '/' + getattr(d['entity'], attr) + d['path']

                if viewType == 'vim.VirtualMachine-Folder' and \
                    type(d['entity']) == vim.VirtualApp and \
                    d['entity'].parentFolder != None:
                    d['entity'] = d['entity'].parentFolder
                    return
                else:
                    d['entity'] = d['entity'].parent
            d['entity'] = d['entity'].parent

        def _get_folder_path():
            while type(d['entity'].parent) != vim.Datacenter:
                d['path'] = '/' + getattr(d['entity'], attr) + d['path']
                d['entity'] = d['entity'].parent
            d['entity'] = d['entity'].parent

        if type(d['entity']) == vim.HostSystem:
            if type(d['entity'].parent) == vim.ClusterComputeResource:
                d['path'] = '/' + getattr(d['entity'], attr)
            d['entity'] = d['entity'].parent
            _get_folder_path()

        elif type(d['entity']) == vim.ResourcePool or type(d['entity']) == vim.VirtualApp:
            _get_resource_pool_path()
            _get_folder_path()

        elif type(d['entity']) == vim.Folder:
            _get_folder_path()

        elif type(d['entity']) == vim.VirtualMachine:
            d['path'] = '/' + getattr(d['entity'], attr)

            if viewType == 'vim.VirtualMachine-Folder':
                if d['entity'].parent == None:
                    d['entity'] = d['entity'].resourcePool
                    _get_resource_pool_path()
                else:
                    d['entity'] = d['entity'].parent
                _get_folder_path()

            elif viewType == 'vim.VirtualMachine-Host':
                if d['entity'].summary.config.template:
                    return d['path']
                d['entity'] = d['entity'].resourcePool
                _get_resource_pool_path()
                _get_folder_path()

        _get_datacenter_path()
        return d['path']

    def listChildEntity(self, parentId='', viewType='', getDetail=False):
        def _parse_data(child):
            if type(child) == vim.VirtualMachine:
                return self._parse_vm_data(child, getDetail, viewCache)
            elif type(child) == vim.Datastore:
                return self._parse_datastore_data(child)
            elif type(child) == vim.Datacenter:
                return {
                    'folder_id': child.vmFolder._moId
                }
            elif type(child) == vim.ComputeResource:
                return {
                    'host_id': child.host[0]._moId,
                    'resourcepool_id': child.resourcePool._moId,
                    'migratable': self.getMigrateCapability(host_obj=child.host[0])
                }
            elif type(child) == vim.ClusterComputeResource:
                return {
                    'resourcepool_id': child.resourcePool._moId
                }
            elif type(child) == vim.HostSystem:
                return {
                    'migratable': self.getMigrateCapability(host_obj=child)
                }
            elif type(child) == vim.VirtualApp:
                return {
                    'started': child.summary.vAppState == 'started'
                }
            elif type(child) == vim.dvs.DistributedVirtualPortgroup:
                return {
                    'type': 'vim.Network'
                }
            else:
                return {}

        children = []
        entities = []
        viewCache = {}

        if parentId == '':
            parentObj = self._get_root_obj()
        else:
            parentObj = self._get_obj_by_moid(parentId)

        if parentId == '':
            if viewType == 'vim.VirtualMachine':
                children = self._get_objs([vim.VirtualMachine], parentObj)
            elif viewType == 'vim.HostSystem':
                children = self._get_objs([vim.HostSystem], parentObj)
            else:
                children = parentObj.childEntity

        elif type(parentObj) == vim.Folder:
            if viewType == 'vim.Folder' and 'VirtualMachine' in parentObj.childType:
                children = [obj for obj in parentObj.childEntity if type(obj) == vim.Folder]
            elif viewType == 'vim.VirtualMachine':
                children = [obj for obj in parentObj.childEntity if type(obj) == vim.VirtualMachine]
            else:
                children = parentObj.childEntity

        elif type(parentObj) == vim.Datacenter:
            if viewType == 'vim.Folder':
                children = [obj for obj in parentObj.vmFolder.childEntity if type(obj) == vim.Folder]
            elif viewType == 'vim.VirtualMachine-Folder':
                children = parentObj.vmFolder.childEntity
            elif viewType == 'vim.ResourcePool':
                children = parentObj.hostFolder.childEntity
            else:
                children = parentObj.hostFolder.childEntity

        elif type(parentObj) == vim.HostSystem:
            if viewType == 'vim.Datastore':
                children = parentObj.datastore
            elif viewType == 'vim.Network':
                if self._is_vcenter():
                    children = [obj for obj in parentObj.network if type(obj) != vim.dvs.DistributedVirtualPortgroup or not obj.config.uplink]
                else:
                    children = [obj for obj in parentObj.network if type(obj) != vim.dvs.DistributedVirtualPortgroup]
            else:
                children = []

        elif type(parentObj) == vim.ClusterComputeResource:
            children = parentObj.host

            if viewType == 'vim.VirtualMachine-Host':
                children += parentObj.resourcePool.resourcePool
                children += parentObj.resourcePool.vm

        elif type(parentObj) == vim.ComputeResource:
            if viewType == 'vim.Datastore':
                children = parentObj.datastore
            elif viewType == 'vim.Network':
                if self._is_vcenter():
                    children = [obj for obj in parentObj.network if type(obj) != vim.dvs.DistributedVirtualPortgroup or not obj.config.uplink]
                else:
                    children = [obj for obj in parentObj.network if type(obj) != vim.dvs.DistributedVirtualPortgroup]
            else:
                if self._is_vcenter():
                    children = parentObj.resourcePool.resourcePool
                    if viewType == 'vim.VirtualMachine-Host':
                        children += parentObj.resourcePool.vm
                else:
                    children = []
                    if viewType == 'vim.VirtualMachine-Host':
                        children += parentObj.host[0].vm

        elif type(parentObj) == vim.ResourcePool or type(parentObj) == vim.VirtualApp:
            children = parentObj.resourcePool

            if viewType == 'vim.VirtualMachine-Folder' or viewType == 'vim.VirtualMachine-Host':
                children += parentObj.vm

        for child in children:
            data = {
                'id': child._moId,
                'name': child.name,
                'type': type(child).__name__
            }
            try:
                entities.append(dict(data, **_parse_data(child)))
            except Exception:
                logger.error(traceback.format_exc())

        entities.sort(key=lambda obj: (obj['type'], obj['name']))
        return entities

    def registerVm(self,
            vm_name=None,
            folder_id=None,
            resourcepool_id=None,
            host_id=None,
            os_name=None,
            cpu_number=None,
            memory_size=None,
            is_system_mbr=None,
            vdisks=None,
            networks=None,
            file_path=None,
            gen_new_uuid=False):

        rp_obj = self._get_obj_by_moid(resourcepool_id, [vim.ResourcePool])
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        folder_obj = self._get_obj_by_moid(folder_id, [vim.Folder])

        def _is_vapp(obj):
            while type(obj) == vim.ResourcePool:
                obj = obj.parent
            return type(obj) == vim.VirtualApp

        try:
            if _is_vapp(rp_obj):
                task = rp_obj.RegisterChildVM_Task(path=file_path, name=vm_name, host=host_obj)
            else:
                task = folder_obj.RegisterVM_Task(path=file_path, name=vm_name, pool=rp_obj, host=host_obj, asTemplate=False)

            vm = self._wait_for_task(task)
            is_create_vm = False

        except (vmodl.fault.InvalidArgument, vim.fault.NotFound):
            config = vim.vm.ConfigSpec(
                name=vm_name,
                guestId=os_name,
                numCPUs=cpu_number,
                memoryMB=memory_size,
                firmware=('bios' if is_system_mbr else 'efi'),
                files=vim.vm.FileInfo(vmPathName=file_path))

            if _is_vapp(rp_obj):
                task = rp_obj.CreateChildVM_Task(config=config, host=host_obj)
            else:
                task = folder_obj.CreateVM_Task(config=config, pool=rp_obj, host=host_obj)

            vm = self._wait_for_task(task)
            is_create_vm = True

        vdisk_idx = 0
        network_idx = 0
        config = vim.vm.ConfigSpec()
        config.deviceChange = []

        try:
            if is_create_vm:
                useIDE = is_system_mbr and (
                    os_name == 'windows7Guest' or
                    os_name == 'windows7_64Guest' or
                    os_name == 'windows7Server64Guest')

                config.deviceChange.append(vim.vm.device.VirtualDeviceSpec(
                    operation=vim.vm.device.VirtualDeviceSpec.Operation.add,
                    device=vim.vm.device.VirtualLsiLogicSASController(sharedBus='noSharing')))

                for vdisk in vdisks:
                    bootable = 'bootable' in vdisk and vdisk['bootable']

                    config.deviceChange.append(vim.vm.device.VirtualDeviceSpec(
                        operation=vim.vm.device.VirtualDeviceSpec.Operation.add,
                        device=vim.vm.device.VirtualDisk(
                            key=-1000000,
                            unitNumber=(0 if useIDE and bootable else vdisk_idx),
                            controllerKey=(200 if useIDE and bootable else config.deviceChange[0].device.key),
                            capacityInKB=vdisk['capacity']/1024,
                            backing=vim.vm.device.VirtualDisk.FlatVer2BackingInfo(
                                diskMode='persistent',
                                fileName=vdisk['filename']))))
                    vdisk_idx += 1

                for network in networks:
                    if network['network_id'].startswith('dvportgroup'):
                        network = self._get_obj_by_key(network['network_id'], [vim.dvs.DistributedVirtualPortgroup])
                        backing = vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo(
                            port=vim.dvs.PortConnection(
                                portgroupKey=network.key,
                                switchUuid=network.config.distributedVirtualSwitch.uuid))
                    else:
                        network = self._get_obj_by_moid(network['network_id'], [vim.Network])
                        backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo(
                            network=network,
                            deviceName=network.name)

                    config.deviceChange.append(vim.vm.device.VirtualDeviceSpec(
                        operation=vim.vm.device.VirtualDeviceSpec.Operation.add,
                        device=vim.vm.device.VirtualE1000(
                            key=-50,
                            backing=backing,
                            addressType='generated',
                            wakeOnLanEnabled=False,
                            connectable=vim.vm.device.VirtualDevice.ConnectInfo(
                                connected=True,
                                startConnected=True,
                                allowGuestControl=True))))
                    network_idx += 1
            else:
                config.instanceUuid = '' if self._is_vcenter() else ' '

                if gen_new_uuid:
                    config.uuid = ' '
                    config.locationId = ' '

                for device in vm.config.hardware.device:
                    spec = vim.vm.device.VirtualDeviceSpec(
                        operation=vim.vm.device.VirtualDeviceSpec.Operation.add,
                        device=eval(type(device).__name__)())

                    if isinstance(device, vim.vm.device.VirtualDisk):
                        vdisk_idx = int(re.findall(r'\d+$', device.deviceInfo.label)[-1]) - 1
                        spec.device.key = -1000000
                        spec.device.unitNumber = device.unitNumber
                        spec.device.controllerKey = device.controllerKey
                        spec.device.capacityInKB = vdisks[vdisk_idx]['capacity']/1024
                        spec.device.backing = vim.vm.device.VirtualDisk.FlatVer2BackingInfo(
                            diskMode=device.backing.diskMode,
                            fileName=vdisks[vdisk_idx]['filename'])

                    elif isinstance(device, vim.vm.device.VirtualEthernetCard):
                        if hasattr(device.backing, 'network') and \
                               (not device.backing.network or \
                                not device.backing.network.summary or \
                                not device.backing.network.summary.accessible):
                            continue
                        if (device.addressType == 'manual' and not gen_new_uuid) or \
                            device.addressType == 'assigned':
                            spec.device.macAddress = device.macAddress
                        elif (device.addressType == 'manual' and gen_new_uuid) or \
                            device.addressType == 'generated':
                            spec.device.macAddress = '00:50:56:%02x:%02x:%02x' % (
                                random.randint(0, 63),
                                random.randint(0, 255),
                                random.randint(0, 255))

                        spec.device.key = -50
                        spec.device.addressType = device.addressType
                        spec.device.connectable = device.connectable
                        spec.device.wakeOnLanEnabled = device.wakeOnLanEnabled

                        if networks[network_idx]['network_id'].startswith('dvportgroup'):
                            network = self._get_obj_by_key(networks[network_idx]['network_id'], [vim.dvs.DistributedVirtualPortgroup])
                            spec.device.backing = vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo(
                                port=vim.dvs.PortConnection(
                                    portgroupKey=network.key,
                                    switchUuid=network.config.distributedVirtualSwitch.uuid))
                        else:
                            network = self._get_obj_by_moid(networks[network_idx]['network_id'], [vim.Network])
                            spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo(
                                network=network,
                                deviceName=network.name)
                        network_idx += 1
                    else:
                        continue

                    config.deviceChange.append(vim.vm.device.VirtualDeviceSpec(
                        operation=vim.vm.device.VirtualDeviceSpec.Operation.remove,
                        device=device))
                    config.deviceChange.append(spec)

            task = vm.ReconfigVM_Task(config)
            self._wait_for_task(task)
        except Exception as e:
            try:
                self.unregisterVm(vm._moId)
            except:
                pass
            raise e

        return vm._moId

    def uploadFile(self, datacenter_id=None, datastore_id=None, src_path=None, dst_path=None):
        dc_obj = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        ds_obj = self._get_obj_by_moid(datastore_id, [vim.Datastore])
        dc_path = self._get_datacenter_path(dc_obj)

        payload = {}

        if version.parse(self.api_version) >= version.parse('6.7') :
            payload['dcPath'] = dc_path
            payload['dsName'] = ds_obj.name
        else:
            payload['dcPath'] = _query_fix(dc_path)
            payload['dsName'] = _query_fix(ds_obj.name)

        url = 'https://%s:%s/folder/%s' % (self.auth_info['hostname'], self.auth_info['port'], _url_fix(dst_path))

        with open(src_path.encode('utf-8'), 'rb') as f:
            resp = requests.put(
                url,
                params=payload,
                verify=False,
                data=f,
                auth=(
                    self.auth_info['username'],
                    self.auth_info['password']))
        if resp.status_code == 401: # retry
            time.sleep(8)
            with open(src_path.encode('utf-8'), 'rb') as f:
                resp = requests.put(_url_fix(url), data=f, auth=(self.auth_info['username'], self.auth_info['password']), verify=False)
        if isSuccessStatusCode(resp.status_code) == False:
            if resp.status_code == 401:
                raise vim.fault.InvalidLogin(msg='Authentication failed when uploading file')
            else:
                raise vim.fault.HostConnectFault(msg='Error while uploading file, status code: %d' % resp.status_code)

    def createDatastore(self, name=None, host_id=None, ip_address=None, mount_path=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        host_datastore_system = host_obj.configManager.datastoreSystem
        volume_spec = vim.host.NasVolume.Specification()
        volume_spec.remoteHost = ip_address
        volume_spec.remotePath = mount_path
        volume_spec.localPath = name
        volume_spec.accessMode = 'readWrite'
        ds_obj = host_datastore_system.CreateNasDatastore(spec=volume_spec)
        return self._parse_datastore_data(ds_obj)

    def removeDatastore(self, host_id=None, datastore_id=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        datastore_obj = self._get_obj_by_moid(datastore_id, [vim.Datastore])
        host_datastore_system = host_obj.configManager.datastoreSystem
        host_datastore_system.RemoveDatastore(datastore=datastore_obj)

    def createFolder(self, datacenter_id=None, folder_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        fileManager.MakeDirectory(name=folder_path, datacenter=datacenter)

    def removeFolder(self, datacenter_id=None, folder_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        task = fileManager.DeleteDatastoreFile_Task(name=folder_path, datacenter=datacenter)
        self._wait_for_task(task)

    def renameFolder(self, datacenter_id=None, src_path=None, dst_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        task = fileManager.MoveDatastoreFile_Task(sourceName=src_path, destinationName=dst_path, sourceDatacenter=datacenter)
        self._wait_for_task(task)

    def createDisk(self, datacenter_id=None, file_path=None, disk_type=None, adapter_type=None, capacity=None):
        spec = vim.VirtualDiskManager.FileBackedVirtualDiskSpec()
        spec.diskType = disk_type
        spec.adapterType = adapter_type
        spec.capacityKb = int(capacity) / 1024
        dc_obj = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        disk_manager = self.si.content.virtualDiskManager
        task = disk_manager.CreateVirtualDisk_Task(name=file_path, datacenter=dc_obj, spec=spec)
        return self._put_task(task)

    def getVmQuestion(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        if vm.runtime.question != None:
            q = self._parse_question(vm.runtime.question)
            q['exist'] = True
            return q
        return {'exist': False}

    def migrateVm(self, vm_id=None, disk_specs=None, datastore_ids=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        relocateSpec = vim.vm.RelocateSpec()

        if not isinstance(disk_specs, list):
            disk_specs = []
        if not isinstance(datastore_ids, list):
            datastore_ids = []

        i = 0
        for device in vm.config.hardware.device:
            if not isinstance(device, vim.vm.device.VirtualDisk):
                continue

            backingInfo = vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
            backingInfo.eagerlyScrub = disk_specs[i]['eager_scrub']
            backingInfo.thinProvisioned = disk_specs[i]['thin_provision']

            datastore = self._get_obj_by_moid(datastore_ids[i], [vim.Datastore])

            diskLocator = vim.vm.RelocateSpec.DiskLocator()
            diskLocator.diskId = device.key
            diskLocator.datastore = datastore
            diskLocator.diskBackingInfo = backingInfo
            relocateSpec.disk.append(diskLocator)
            i += 1

        relocateSpec.datastore = relocateSpec.disk[0].datastore

        task = vm.RelocateVM_Task(spec=relocateSpec)
        return self._put_task(task)

    def poweronVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        poweron_task = vm.PowerOn()
        time.sleep(10)

        if vm.runtime.question != None:
            question = self._parse_question(vm.runtime.question)
            vm.AnswerVM(question['question_id'], question['default_answer'])
            time.sleep(5)

        try:
            self._wait_for_task(poweron_task)
        except vim.fault.InvalidPowerState as e:
            if e.existingState != 'poweredOn':
                raise e

    def poweroffVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        poweroff_task = vm.PowerOff()
        try:
            self._wait_for_task(poweroff_task)
        except vim.fault.InvalidPowerState as e:
            if e.existingState != 'poweredOff':
                raise e

    def unregisterVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        vm.UnregisterVM()

    def listSnapshot(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        if hasattr(vm, 'snapshot') and hasattr(vm.snapshot, 'rootSnapshotList'):
            return self._list_snapshot_recursively(vm.snapshot.rootSnapshotList)
        else:
            return []

    def getSnapshot(self, vm_id=None, snap_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
        return self._parse_snapshot_data(snap_obj)

    def takeSnapshot(self, vm_id=None, snap_name=None, snap_desc=None, quiesce=False):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        create_snap_task = vm.CreateSnapshot_Task(
            name=snap_name,
            description=snap_desc,
            memory=False,
            quiesce=quiesce)
        snap_obj = self._wait_for_task(create_snap_task)
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_obj._moId)
        return self._parse_snapshot_data(snap_obj)

    def removeSnapshot(self, vm_id=None, snap_id=None, nonblocking=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
        task = snap_obj.snapshot.RemoveSnapshot_Task(
                removeChildren=False,
                consolidate=True)
        if not nonblocking:
            self._wait_for_task(task)

    def getTaskInfo(self, task_id=None):
        task = self._get_task(task_id)

        if task.info.state == 'error':
            self._del_task(task_id)
            raise task.info.error

        if task.info.state == 'success':
            self._del_task(task_id)

        return {
            'success': task.info.state == 'success',
            'progress': task.info.progress,
            'result': task.info.result
        }

    def cancelTask(self, task_id=None):
        task = self._get_task(task_id)
        task.CancelTask()
        self._del_task(task_id)

    def getCbtState(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        return {
            'support': vm.capability.changeTrackingSupported,
            'enabled': vm.config.changeTrackingEnabled
        }

    def setCbtState(self, vm_id=None, enable=False):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])

        if enable == vm.config.changeTrackingEnabled:
            return None

        cspec = vim.vm.ConfigSpec()
        cspec.changeTrackingEnabled = enable
        reconfig_vm_task = vm.ReconfigVM_Task(cspec)
        self._wait_for_task(reconfig_vm_task)

        snap_name = SNAPSHOT_NAME_PREFIX + str(datetime.datetime.now())
        snap = self.takeSnapshot(vm_id, snap_name, SNAPSHOT_DESC, False)
        self.removeSnapshot(vm_id, snap['snap_id'])

    def getDiskCbt(self, vm_id=None, snap_id=None, key=None, offset=None, change_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
        cbt_obj = vm.QueryChangedDiskAreas(
            snapshot=snap_obj.snapshot,
            deviceKey=int(key),
            startOffset=offset,
            changeId=change_id)

        out_json = {}
        out_json['range_length'] = cbt_obj.length
        out_json['change_areas'] = []

        for extent in cbt_obj.changedArea:
            out_json['change_areas'].append({'offset': extent.start, 'length': extent.length})

        return out_json

    def updataMetaInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, is_folder=False, isWindows=False):

        def guestUpdateFileMeta(vm, vm_creds, file_attribute, msg):
            try:
                self.si.content.guestOperationsManager.fileManager. \
                        ChangeFileAttributesInGuest(vm, vm_creds, dst_path, file_attribute)
            except Exception as e:
                raise Exception('%s[%s] error[%s]' % (msg,dst_path, str(e)))

        def updateMeta(vm, vm_creds, org_attribute, src_stat, attrList):
            file_attribute = copy.deepcopy(org_attribute)
            if ENUM_ATTR_MTIME in attrList:
                file_attribute.modificationTime = datetime.datetime.fromtimestamp(src_stat.st_mtime)
            if ENUM_ATTR_CTIME in attrList:
                file_attribute.createTime = datetime.datetime.fromtimestamp(src_stat.st_ctime)
            if ENUM_ATTR_ATIME in attrList:
                file_attribute.accessTime = datetime.datetime.fromtimestamp(src_stat.st_atime)
            if ENUM_ATTR_PERMS in attrList:
                file_attribute.permissions = src_stat.st_mode
            if ENUM_ATTR_UID in attrList:
                file_attribute.ownerId = src_stat.st_uid
            if ENUM_ATTR_GID in attrList:
                file_attribute.groupId = src_stat.st_gid
            guestUpdateFileMeta(vm, vm_creds, file_attribute, 'update [%s]' % (', '.join(str(x) for x in attrList)))

        def getCanUpdateAttr(file_attribute):
            attrList = [ENUM_ATTR_ATIME, ENUM_ATTR_MTIME]
            if hasattr(file_attribute, 'permissions'):
                attrList.append(ENUM_ATTR_PERMS)
            if vm_user == 'root':
                if hasattr(file_attribute, 'ownerId'):
                    attrList.append(ENUM_ATTR_UID)
                if hasattr(file_attribute, 'groupId'):
                    attrList.append(ENUM_ATTR_GID)
            return attrList

        filesInfo = None
        src_stat = None
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        try:
            src_stat = os.lstat(src_path.encode('utf-8'))
        except Exception as e:
            raise LocalStatFail(-1, 'failed to stat[%s] error[%s]' % (msg,dst_path, str(e)))

        if is_folder:
            filesInfo = self.si.content.guestOperationsManager.fileManager. \
                    ListFilesInGuest(vm, creds, dst_path+'/', 0, 1)
        else:
            filesInfo = self.si.content.guestOperationsManager.fileManager. \
                    ListFilesInGuest(vm, creds, dst_path, maxResults = 1)
        if len(filesInfo.files) != 1:
            raise GuestOSMetaControlFail(None, -1, 'Invaild file Path[%s]' % (dst_path), None)

        file_attribute = filesInfo.files[0].attributes
        attr_list = getCanUpdateAttr(file_attribute)

        try:
            updateMeta(vm, creds, file_attribute, src_stat, attr_list)
            return
        except Exception as e:
            logger.error('update [%s] attribute [%s] failed, error[%s]' % (dst_path, ', '.join(str(x) for x in attr_list), str(e)))

        fixed_attr_list = []
        miss_attr_list = []
        for enum_attr in attr_list:
            fixed_attr_list.append(enum_attr)
            try:
                updateMeta(vm, creds, file_attribute, src_stat, fixed_attr_list)
            except Exception as e:
                logger.error('miss [%d] error[%s]' % (enum_attr, str(e)))
                fixed_attr_list.pop()
                miss_attr_list.append(enum_attr)

        if len(miss_attr_list) > 0:
            raise GuestOSMetaControlFail(None, -1, 'update [%s] attribute [%s] failed' % (dst_path, ', '.join(str(x) for x in miss_attr_list)), miss_attr_list)
        return

    def uploadFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, overwrite=True):
        def uploadFileInVmTask(task):
            class LengthWrapper:
                def __init__(self, readable, length, task):
                    self._length = self._remaining = length
                    self._readable = readable
                    self._task = task

                def read(self, *args, **kwargs):
                    if self._task._canceled:
                        raise Exception('cancel')
                    if self._remaining <= 0:
                        return ''
                    chunk = self._readable.read(*args, **kwargs)[:self._remaining]
                    self._remaining -= len(chunk)
                    self._task.info.progress = 100.0 * (self._length - self._remaining) / self._length
                    return chunk

                def __len__(self):
                    return self._length

            try:
                vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
                creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)

                file_size = os.path.getsize(src_path.encode('utf-8'))
                file_attribute = vim.vm.guest.FileManager.FileAttributes()
                temp_dst_path = dst_path + '_abb.' + randomString()

                # check dest. file available
                task.info.result = False
                self.si.content.guestOperationsManager.fileManager. \
                        InitiateFileTransferToGuest(vm, creds, dst_path, file_attribute, file_size, overwrite)
                # if file exist and not overwrite, InitiateFileTransferToGuest may throw exception
                task.info.result = True

                url = self.si.content.guestOperationsManager.fileManager. \
                        InitiateFileTransferToGuest(vm, creds, temp_dst_path, file_attribute, file_size, overwrite). \
                        replace('*', self.auth_info['hostname'], 1)

                with open(src_path.encode('utf-8'), 'rb') as file:
                    lw = LengthWrapper(file, file_size, task)
                    resp = requests.put(url, data=lw, verify=False, timeout=900)
                if isSuccessStatusCode(resp.status_code):
                    task.info.state = 'success'
                    self.si.content.guestOperationsManager.fileManager.MoveFileInGuest(vm, creds, temp_dst_path, dst_path, overwrite)
                elif resp.status_code == 401:
                    raise vim.fault.InvalidLogin(msg='Authentication failed when uploading file in VM')
                else:
                    raise vim.fault.HostConnectFault(msg='Error while uploading file, status code: %d' % resp.status_code)
            except Exception as e:
                if type(e) == vim.fault.FileAlreadyExists:
                    task.info.state = 'success'
                elif type(e) == vim.fault.NotAFile:
                    task.info.state = 'error'
                    task.info.error = GuestOSException(None, -1, 'Failed to overwrite file/folder. Please make sure the file/folder property is the same.')
                else:
                    task.info.state = 'error'
                    task.info.error = e
                    if type(e) == requests.ConnectionError and hasattr(e.args[0][1], 'errno') and e.args[0][1].errno == errno.EPIPE:
                        task.info.error = GuestNotEnoughSpace('No enough space to restore file.')
                try:
                    self.si.content.guestOperationsManager.fileManager.DeleteFileInGuest(vm, creds, temp_dst_path)
                except:
                    pass

        class FakeTaskInfo:
            def __init__(self):
                self.key = 'upload_to_vm_' + uuid.uuid4().hex[:12]
                self.state = 'running'
                self.error = None
                self.result = None
                self.progress = 0

        class FakeTask:
            def __init__(self):
                self.info = FakeTaskInfo()
                self._canceled = False
            def CancelTask(self):
                self._canceled = True

        task = FakeTask()
        threading.Thread(target=uploadFileInVmTask, args=(task,)).start()
        return self._put_task(task)

    def listFilesInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, index=None, max_results=None, match_pattern=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        auth = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        authMgr = self.si.content.guestOperationsManager.authManager
        fileMgr = self.si.content.guestOperationsManager.fileManager
        out = {'files': [], 'remaining': 0}

        if file_path == '':
            authMgr.ValidateCredentialsInGuest(vm, auth)
            for disk in vm.guest.disk:
                out['files'].append({
                    'name': disk.diskPath.replace('\\', ''),
                    'path': disk.diskPath.replace('\\', ''),
                    'size': 0,
                    'is_dir': True,
                    'mtime': 0
                })
        else:
            resp = fileMgr.ListFilesInGuest(vm, auth, file_path, index, max_results, match_pattern)
            out['remaining'] = resp.remaining
            for file in resp.files:
                if file.path == '.' or file.path == '..':
                    continue
                out['files'].append({
                    'name': file.path,
                    'path': file_path + '/' + file.path,
                    'size': file.size,
                    'is_dir': file.type == 'directory',
                    'mtime': int(file.attributes.modificationTime.strftime('%s'))
                })

        return out

    def downloadFileFromVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        fileMgr = self.si.content.guestOperationsManager.fileManager
        fileInfo = fileMgr.InitiateFileTransferFromGuest(vm, creds, src_path)
        url = fileInfo.url.replace('*', self.auth_info['hostname'], 1)
        resp = requests.get(url, stream=True, verify=False)

        if isSuccessStatusCode(resp.status_code) == False:
            if resp.status_code == 401:
                raise vim.fault.InvalidLogin(msg='Authentication failed when downloading file from VM')
            else:
                raise vim.fault.HostConnectFault(msg='Error while downloading file, status code: %d' % resp.status_code)
        with open(dst_path.encode('utf-8'), 'wb') as myfile:
            myfile.write(resp.text.encode('utf-8'))

    def removeFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        self.si.content.guestOperationsManager.fileManager.DeleteFileInGuest(vm, creds, file_path)

    def readEnvVariableInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, names=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        results = self.si.content.guestOperationsManager.processManager.ReadEnvironmentVariableInGuest(vm, creds, names)

        variables = []
        for result in results:
            variables.append(result)
        return variables

    def executeProgramInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, arguments=None, timeout=180):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        pm = self.si.content.guestOperationsManager.processManager
        ps = vim.vm.guest.ProcessManager.ProgramSpec(programPath=file_path, arguments=arguments)
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        res = pm.StartProgramInGuest(vm, creds, ps)
        execute_interval = 0
        pid_exitcode = -1

        if res > 0:
            pid_exitcode = pm.ListProcessesInGuest(vm, creds, [res]).pop().exitCode
            while (re.match('[^0-9]+', str(pid_exitcode))):
                if (execute_interval >= timeout):
                    break
                time.sleep(5)
                execute_interval += 5
                pid_exitcode = pm.ListProcessesInGuest(vm, creds, [res]).pop().exitCode
                if (re.match('[0-9]+', str(pid_exitcode))):
                    break

        out_json = {}
        out_json['exit_code'] = pid_exitcode
        return out_json

    def createDirectoryInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, dir_path=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        try:
            self.si.content.guestOperationsManager.fileManager.MakeDirectoryInGuest(vm, creds, dir_path, True)
        except vim.fault.FileAlreadyExists:
            try:
                self.si.content.guestOperationsManager.fileManager.ListFilesInGuest(vm, creds, dir_path+'/', 0, 1)
            except vmodl.fault.InvalidArgument:
                raise GuestOSException(None, -1, 'Failed to overwrite file/folder. Please make sure the file/folder property is the same.')
            return False
        return True

    def downloadFile(self, vm_id=None, dst_path=None, src_path=None):
        vm_obj = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        dc_obj = self._get_datacenter_obj(vm_obj)
        dc_path = self._get_datacenter_path(dc_obj)
        ds_name = src_path.split(']')[0][1:]
        vm_path = src_path[src_path.find(']') + 2:]

        payload = {}

        if version.parse(self.api_version) >= version.parse('6.7') :
            payload['dcPath'] = dc_path
            payload['dsName'] = ds_name
        else:
            payload['dcPath'] = _query_fix(dc_path)
            payload['dsName'] = _query_fix(ds_name)

        url = 'https://%s:%s/folder/%s' % (self.auth_info['hostname'], self.auth_info['port'], _url_fix(vm_path))

        for retryCount in range(10):
            try:
                resp = requests.get(
                    url,
                    params=payload,
                    verify=False,
                    auth=(
                        self.auth_info['username'],
                        self.auth_info['password']))
                if resp.status_code == 401: # retry
                    time.sleep(8)
                    resp = requests.get(
                        url,
                        params=payload,
                        verify=False,
                        auth=(
                            self.auth_info['username'],
                            self.auth_info['password']))
                break
            except (requests.ConnectionError, requests.Timeout, requests.ConnectTimeout, requests.ReadTimeout):
                logger.error('get connection error! retryCount = %d' % retryCount)
                time.sleep(2**retryCount)

        if isSuccessStatusCode(resp.status_code) == False:
            if resp.status_code == 401:
                raise vim.fault.InvalidLogin(msg='Authentication failed when downloading file')
            elif resp.status_code == 404:
                raise vim.fault.VimFault(msg='File not found: "%s"' % src_path)
            elif resp.status_code == 503:
                raise vim.fault.VimFault(msg='Downloading "%s" from datastore is not available.' % src_path)
            else:
                raise vim.fault.HostConnectFault(msg='Error while downloading file, status code: %d' % resp.status_code)

        with open(dst_path.encode('utf-8'), 'wb') as file:
            file.write(resp.content)

    def getMigrateCapability(self, host_id=None, host_obj=None):
        if host_obj == None:
            host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        return {
            'vmotion': host_obj.capability.storageVMotionSupported,
            'vcenter': self._is_vcenter()
        }

    def getHostInterface(self, host_id=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        return [vnic.spec.ip.ipAddress for vnic in host_obj.configManager.vmotionSystem.netConfig.candidateVnic]

    def runShell(self, command, timeout):
        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(self.auth_info['hostname'],
                username=self.auth_info['username'],
                password=self.auth_info['password'])

            stdin, stdout, stderr = ssh.exec_command(command)
        except Exception:
            raise vim.fault.HostConnectFault(msg='SSH connection failed.')

        if timeout > 0:
            elapsed = 0
            while not stdout.channel.exit_status_ready():
                if elapsed >= timeout:
                   raise vim.fault.Timedout(msg='execute [%s] timeout' % command)
                time.sleep(1)
                elapsed += 1

        status = stdout.channel.recv_exit_status()
        stdout = stdout.read().strip('\n')
        stderr = stderr.read().strip('\n')
        ssh.close()

        if status != 0:
            m = re.search(r'\((.*?)\) \{.*?msg = "(.*?)".*?\}', stderr, re.DOTALL)
            if m != None and m.lastindex == 2:
                raise eval(m.group(1))(msg=m.group(2))
            if 0 <= stderr.find('File exists'):
                raise vim.fault.FileAlreadyExists(msg=stderr)
            raise vim.fault.VimFault(msg=stderr)
        return stdout

    def listSession(self):
        logger.info('session list of %s' % self.auth_info['hostname'])
        sessionList = self.si.content.sessionManager.sessionList
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(('8.8.8.8', 80))
        hostAddress = sock.getsockname()[0]
        sock.close()

        for session in sessionList:
            if session.ipAddress == hostAddress:
                logger.info(session.key, session.loginTime, session.userAgent)

    ####  util functions ###
    def _is_vcenter(self):
        return self.api_type == 'VirtualCenter'

    def _get_root_obj(self):
        return self.si.content.rootFolder

    def _get_obj(self, key, value, vimType=None, startObj=None):
        objs = self._get_objs(vimType, startObj)
        for obj in objs:
            if getattr(obj, key) == value:
                return obj
        raise vim.fault.NotFound(msg='Cannot find [%s]=[%s]' % (key, value))

    def _get_obj_by_moid(self, moId, vimType=None, startObj=None):
        return self._get_obj('_moId', moId, vimType, startObj)

    def _get_obj_by_key(self, key, vimType=None, startObj=None):
        return self._get_obj('key', key, vimType, startObj)

    def _get_objs(self, vimType=None, startObj=None):
        if startObj is None:
            startObj = self._get_root_obj()

        container = self.si.content.viewManager.CreateContainerView(startObj, vimType, True)
        objs = [obj for obj in container.view]
        container.Destroy()
        return objs

    def _get_snapshot_obj(self, snapshots, moId):
        snapshot_obj_data = self._list_snapshot_obj_recursively(snapshots)
        for obj in snapshot_obj_data:
            if obj.snapshot._moId == moId:
                return obj
        raise vim.fault.NotFound(msg='Cannot find [%s]' % moId)

    def _list_snapshot_obj_recursively(self, snapshots):
        snapshot_obj_data = []
        for snapshot in snapshots:
            snapshot_obj_data.append(snapshot)
            snapshot_obj_data += self._list_snapshot_obj_recursively(snapshot.childSnapshotList)
        return snapshot_obj_data

    def _list_snapshot_recursively(self, snapshots):
        snapshot_data = []
        for snapshot in snapshots:
            snapshot_data.append(self._parse_snapshot_data(snapshot))
            snapshot_data += self._list_snapshot_recursively(snapshot.childSnapshotList)
        return snapshot_data

    def _parse_snapshot_data(self, snapshot):
        obj = {}
        obj['snap_id'] = snapshot.snapshot._moId
        obj['name'] = snapshot.name
        obj['desc'] = snapshot.description
        obj['cbt_enable'] = snapshot.snapshot.config.changeTrackingEnabled
        obj['create_time'] = int(time.mktime(snapshot.createTime.timetuple()))
        obj['config_files'] = self._get_config_files(snapshot.vm)
        obj['virtual_disk'] = self._get_virtual_disk(snapshot.snapshot.config.hardware.device)
        return obj

    def _parse_question(self, question):
        choice = question.choice
        defaultIndex = (choice.defaultIndex if choice.defaultIndex else 0)
        defaultAnswer = choice.choiceInfo[defaultIndex]

        if defaultAnswer.label == 'button.uuid.copiedTheVM':
            for i in range(len(choice.choiceInfo)):
                if choice.choiceInfo[i].label == 'button.uuid.movedTheVM':
                    defaultAnswer = choice.choiceInfo[i]
                    break

        return {'question_id': question.id, 'default_answer': defaultAnswer.key}

    def _parse_datastore_data(self, datastore):
        info = datastore.info
        summary = datastore.summary

        out = {}
        out['name'] = summary.name
        out['accessible'] = summary.accessible
        out['capacity'] = summary.capacity
        out['free_space'] = summary.freeSpace
        out['datastore_id'] = datastore._moId

        if hasattr(info, 'nas'):
            out['volume_type'] = info.nas.type
            out['remote_host'] = info.nas.remoteHost
            out['remote_path'] = info.nas.remotePath
        if hasattr(info, 'vmfs'):
            out['volume_type'] = info.vmfs.type
        return out

    def _parse_vm_data(self, vm, get_detail=False, view_cache={}):
        out = {}
        summary = vm.summary
        guest = summary.guest
        config = summary.config
        runtime = summary.runtime
        storage = summary.storage
        host = runtime.host

        out['vm_id'] = vm._moId
        out['name'] = config.name
        out['os_name'] = config.guestFullName
        out['is_template'] = config.template
        out['instance_uuid'] = config.instanceUuid
        out['ram_mb'] = config.memorySizeMB
        out['num_cpu'] = config.numCpu
        out['powered_on'] = runtime.powerState == vim.VirtualMachinePowerState.poweredOn
        out['vmtool_running'] = guest.toolsRunningStatus == vim.VirtualMachineToolsRunningStatus.guestToolsRunning
        out['status'] = runtime.connectionState
        out['status_ok'] = runtime.connectionState == 'connected'

        if get_detail and vm.config:
            if view_cache.get(host._moId) == None:
                view_cache[host._moId] = host.name if self._is_vcenter() else ''
            out['hypervisor_name'] = view_cache[host._moId]

            if storage:
                out['storage_size'] = storage.committed

            folder = vm.parent if vm.parent != None else vm.resourcePool
            if view_cache.get(folder._moId) == None:
                view_cache[folder._moId] = self.getEntityPath(entity=folder, viewType='vim.VirtualMachine-Folder', attr='name')
            out['folder_path'] = view_cache[folder._moId]

            out['virtual_disk'] = self._get_virtual_disk(vm.config.hardware.device)
            out['config_files'] = self._get_config_files(vm)

            dc = self._get_datacenter_obj(vm)
            if vm.parent:
                folder = vm.parent
                out['folder'] = {
                    'name': dc.name if folder.parent == dc else folder.name,
                    'folder_id': folder._moId,
                    'datacenter_id': dc._moId
                }
            else:
                out['folder'] = {
                    'name': dc.name,
                    'folder_id': dc.vmFolder._moId,
                    'datacenter_id': dc._moId
                }
            if not out['is_template']:
                rp = vm.resourcePool
                out['resourcepool'] = {
                    'name': rp.parent.name if isinstance(rp.parent, vim.ComputeResource) else rp.name,
                    'is_vapp': (vm.parentVApp != None),
                    'resourcepool_id': rp._moId
                }
            out['hypervisor'] = {
                'name': host.name,
                'host_id': host._moId,
                'standalone': type(host.parent) == vim.ComputeResource
            }
            out['datastore'] = []
            for device in vm.config.hardware.device:
                if not isinstance(device, vim.vm.device.VirtualDisk):
                    continue
                datastore = device.backing.datastore
                if not datastore:
                    continue
                out['datastore'].append({
                    'name': datastore.summary.name,
                    'datastore_id': datastore._moId
                })
            out['network'] = []
            for device in vm.config.hardware.device:
                if not isinstance(device, vim.vm.device.VirtualEthernetCard):
                    continue
                backing = device.backing
                if isinstance(backing, vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo):
                    try:
                        network = self._get_obj_by_key(backing.port.portgroupKey, [vim.dvs.DistributedVirtualPortgroup])
                    except vim.fault.NotFound:
                        continue
                else:
                    network = backing.network

                if network is None:
                    continue

                out['network'].append({
                    'name': network.summary.name,
                    'network_id': network._moId
                })
            out['is_bios'] = vm.config.firmware == 'bios'
            out['hw_version'] = vm.config.version.split('-')[1]
            out['cbt_enabled'] = vm.config.changeTrackingEnabled

        return out

    def _get_config_files(self, vm):
        files = []
        if not hasattr(vm, 'layoutEx') or not hasattr(vm.layoutEx, 'file'):
            return files
        for file in vm.layoutEx.file:
            if file.name.endswith('vmx') or file.name.endswith('vmxf') or file.name.endswith('nvram'):
                files.append(file.name)
        return files

    def _get_virtual_disk(self, devices):
        def _get_virtual_disk_filename_root(backing):
            while backing.parent:
                backing = backing.parent
            return backing.fileName

        disks = []
        devices = { device.key:device for device in devices }

        for device in devices.itervalues():
            if not isinstance(device, vim.vm.device.VirtualDisk):
                continue
            backing = device.backing
            if not backing.datastore:
                continue

            disk = {}
            disk['key'] = device.key
            disk['name'] = device.deviceInfo.label
            disk['unit_number'] = device.unitNumber
            disk['capacity'] = device.capacityInKB * 1024
            disk['uuid'] = backing.uuid
            disk['filename'] = backing.fileName
            disk['change_id'] = backing.changeId
            disk['eager_scrub'] = backing.eagerlyScrub if hasattr(backing, 'eagerlyScrub') else False
            disk['thin_provision'] = backing.thinProvisioned if hasattr(backing, 'thinProvisioned') else False
            disk['filename_root'] = _get_virtual_disk_filename_root(backing)
            disk['disk_mode'] = backing.diskMode
            disk['disk_type'] = type(backing).__name__
            controller = devices[device.controllerKey]
            disk['bus_number'] = controller.busNumber
            disk['controller_type'] = type(controller).__name__
            disks.append(disk)

        disks.sort(key=lambda x:int(re.findall(r'\d+$', x['name'])[-1]))
        return disks

    def _get_datacenter_obj(self, obj):
        while type(obj) != vim.Datacenter:
            if hasattr(obj, 'parent') and obj.parent:
                obj = obj.parent
            if hasattr(obj, 'parentVApp') and obj.parentVApp:
                obj = obj.parentVApp
        return obj

    def _get_datacenter_path(self, obj):
        paths = []
        while obj != self._get_root_obj():
            paths.append(obj.name)
            obj = obj.parent
        return '/'.join(reversed(paths))

    def _wait_for_task(self, task, timeout=2147483648):
        now = time.time()

        while True:
            state = task.info.state
            if (time.time() - now) > timeout:
                task.CancelTask()
                raise vim.fault.Timedout(msg='task timeout. task id [%s]' % task.info.key)
            if state == 'success':
                break
            elif state == 'error':
                raise task.info.error
            elif state in ('queued', 'running'):
                pass
            else:
                task.CancelTask()
                raise vim.fault.InvalidState(msg='unknown task state [%s]' % state)
            time.sleep(1)

        return task.info.result

    def _get_task(self, task_id):
        return self.task_pool[task_id]

    def _put_task(self, task):
        self.task_pool[task.info.key] = task
        return task.info.key

    def _del_task(self, task_id):
        del self.task_pool[task_id]
